@@ -1,9 +1,10 @@
-/* Copyright (c) 2004-2010 Dovecot authors, see the included COPYING file */
+/* Copyright (c) 2004-2011 Dovecot authors, see the included COPYING file */
 
-#include "common.h"
+#include "auth-common.h"
 
 #if defined(PASSDB_CHECKPASSWORD) || defined(USERDB_CHECKPASSWORD)
 
+#include "var-expand.h"
 #include "db-checkpassword.h"
 
 static void env_put_extra_fields(const char *extra_fields)
@@ -129,6 +130,18 @@ void checkpassword_setup_env(struct auth_request *request)
 	}
 }
 
+const char *
+checkpassword_get_cmd(struct auth_request *request, const char *args,
+		      const char *checkpassword_reply_path)
+{
+	string_t *str;
+
+	str = t_str_new(256);
+	var_expand(str, args,
+		   auth_request_get_var_expand_table(request, NULL));
+	return t_strconcat(str_c(str), " ", checkpassword_reply_path, NULL);
+}
+
 void checkpassword_child_input(struct chkpw_auth_request *request)
 {
 	unsigned char buf[1024];
