@@ -1,9 +1,9 @@
-/* Copyright (c) 2009-2010 Dovecot authors, see the included COPYING file */
+/* Copyright (c) 2009-2011 Dovecot authors, see the included COPYING file */
 
-#include "common.h"
-#include "mech.h"
+#include "auth-common.h"
 #include "passdb.h"
-#include "plain-common.h"
+#include "mech.h"
+#include "mech-plain-common.h"
 
 static void
 mech_external_auth_continue(struct auth_request *request,
@@ -42,8 +42,8 @@ static struct auth_request *mech_external_auth_new(void)
 const struct mech_module mech_external = {
 	"EXTERNAL",
 
-	MEMBER(flags) 0,
-	MEMBER(passdb_need) MECH_PASSDB_NEED_VERIFY_PLAIN,
+	.flags = 0,
+	.passdb_need = MECH_PASSDB_NEED_VERIFY_PLAIN,
 
 	mech_external_auth_new,
 	mech_generic_auth_initial,
