@@ -1,10 +1,11 @@
-/* Copyright (c) 2002-2010 Dovecot authors, see the included COPYING file */
+/* Copyright (c) 2002-2011 Dovecot authors, see the included COPYING file */
 
-#include "common.h"
+#include "imap-common.h"
 #include "str.h"
 #include "ostream.h"
 #include "imap-resp-code.h"
-#include "commands.h"
+#include "imap-util.h"
+#include "imap-commands.h"
 #include "imap-search-args.h"
 
 #include <time.h>
@@ -27,7 +28,7 @@ static void client_send_sendalive_if_needed(struct client *client)
 	}
 }
 
-static int fetch_and_copy(struct client *client, struct mailbox *destbox,
+static int fetch_and_copy(struct client *client,
 			  struct mailbox_transaction_context *t,
 			  struct mail_search_args *search_args,
 			  const char **src_uidset_r,
@@ -36,8 +37,6 @@ static int fetch_and_copy(struct client *client, struct mailbox *destbox,
 	struct mail_search_context *search_ctx;
         struct mailbox_transaction_context *src_trans;
 	struct mail_save_context *save_ctx;
-	struct mail_keywords *keywords;
-	const char *const *keywords_list;
 	struct mail *mail;
 	unsigned int copy_count = 0;
 	struct msgset_generator_context srcset_ctx;
@@ -50,10 +49,9 @@ static int fetch_and_copy(struct client *client, struct mailbox *destbox,
 	src_trans = mailbox_transaction_begin(client->mailbox, 0);
 	search_ctx = mailbox_search_init(src_trans, search_args, NULL);
 
-	mail = mail_alloc(src_trans, MAIL_FETCH_STREAM_HEADER |
-			  MAIL_FETCH_STREAM_BODY, NULL);
+	mail = mail_alloc(src_trans, 0, NULL);
 	ret = 1;
-	while (mailbox_search_next(search_ctx, mail) > 0 && ret > 0) {
+	while (mailbox_search_next(search_ctx, mail) && ret > 0) {
 		if (mail->expunged) {
 			ret = 0;
 			break;
@@ -62,17 +60,11 @@ static int fetch_and_copy(struct client *client, struct mailbox *destbox,
 		if ((++copy_count % COPY_CHECK_INTERVAL) == 0)
 			client_send_sendalive_if_needed(client);
 
-		keywords_list = mail_get_keywords(mail);
-		keywords = str_array_length(keywords_list) == 0 ? NULL :
-			mailbox_keywords_create_valid(destbox, keywords_list);
-
 		save_ctx = mailbox_save_alloc(t);
-		mailbox_save_set_flags(save_ctx, mail_get_flags(mail),
-				       keywords);
+		mailbox_save_copy_flags(save_ctx, mail);
 
 		if (mailbox_copy(&save_ctx, mail) < 0)
 			ret = mail->expunged ? 0 : -1;
-		mailbox_keywords_free(destbox, &keywords);
 
 		msgset_generator_next(&srcset_ctx, mail->uid);
 	}
@@ -93,15 +85,18 @@ static int fetch_and_copy(struct client *client, struct mailbox *destbox,
 bool cmd_copy(struct client_command_context *cmd)
 {
 	struct client *client = cmd->client;
-	struct mail_storage *storage;
+	struct mail_namespace *dest_ns;
+	struct mail_storage *dest_storage;
 	struct mailbox *destbox;
 	struct mailbox_transaction_context *t;
         struct mail_search_args *search_args;
-	const char *messageset, *mailbox, *src_uidset, *msg = NULL;
+	const char *messageset, *mailbox, *storage_name, *src_uidset;
+	enum mailbox_name_status status;
 	enum mailbox_sync_flags sync_flags = 0;
 	enum imap_sync_flags imap_flags = 0;
+	struct mail_transaction_commit_changes changes;
 	unsigned int copy_count;
-	uint32_t uid_validity, uid1, uid2;
+	string_t *msg;
 	int ret;
 
 	/* <message set> <mailbox> */
@@ -111,28 +106,39 @@ bool cmd_copy(struct client_command_context *cmd)
 	if (!client_verify_open_mailbox(cmd))
 		return TRUE;
 
-	/* open the destination mailbox */
-	if (!client_verify_mailbox_name(cmd, mailbox,
-				CLIENT_VERIFY_MAILBOX_SHOULD_EXIST_TRYCREATE))
-		return TRUE;
-
 	ret = imap_search_get_seqset(cmd, messageset, cmd->uid, &search_args);
 	if (ret <= 0)
 		return ret < 0;
 
-	storage = client_find_storage(cmd, &mailbox);
-	if (storage == NULL)
+	/* open the destination mailbox */
+	dest_ns = client_find_namespace(cmd, mailbox, &storage_name, &status);
+	if (dest_ns == NULL)
 		return TRUE;
 
-	if (mailbox_equals(client->mailbox, storage, mailbox))
+	switch (status) {
+	case MAILBOX_NAME_EXISTS_MAILBOX:
+		break;
+	case MAILBOX_NAME_EXISTS_DIR:
+		status = MAILBOX_NAME_VALID;
+		/* fall through */
+	case MAILBOX_NAME_VALID:
+	case MAILBOX_NAME_INVALID:
+	case MAILBOX_NAME_NOINFERIORS:
+		client_fail_mailbox_name_status(cmd, mailbox,
+						"TRYCREATE", status);
+		return TRUE;
+	}
+
+	if (mailbox_equals(client->mailbox, dest_ns, storage_name))
 		destbox = client->mailbox;
 	else {
-		destbox = mailbox_open(&storage, mailbox, NULL,
-				       MAILBOX_OPEN_SAVEONLY |
-				       MAILBOX_OPEN_FAST |
-				       MAILBOX_OPEN_KEEP_RECENT);
-		if (destbox == NULL) {
-			client_send_storage_error(cmd, storage);
+		destbox = mailbox_alloc(dest_ns->list, storage_name,
+					MAILBOX_FLAG_SAVEONLY |
+					MAILBOX_FLAG_KEEP_RECENT);
+		if (mailbox_open(destbox) < 0) {
+			client_send_storage_error(cmd,
+				mailbox_get_storage(destbox));
+			mailbox_free(&destbox);
 			return TRUE;
 		}
 		if (client->enabled_features != 0)
@@ -142,49 +148,47 @@ bool cmd_copy(struct client_command_context *cmd)
 	t = mailbox_transaction_begin(destbox,
 				      MAILBOX_TRANSACTION_FLAG_EXTERNAL |
 				      MAILBOX_TRANSACTION_FLAG_ASSIGN_UIDS);
-	ret = fetch_and_copy(client, destbox, t, search_args,
-			     &src_uidset, &copy_count);
+	ret = fetch_and_copy(client, t, search_args, &src_uidset, &copy_count);
 	mail_search_args_unref(&search_args);
 
+	msg = t_str_new(256);
 	if (ret <= 0)
 		mailbox_transaction_rollback(&t);
-	else if (mailbox_transaction_commit_get_uids(&t, &uid_validity,
-						     &uid1, &uid2) < 0)
+	else if (mailbox_transaction_commit_get_changes(&t, &changes) < 0)
 		ret = -1;
-	else if (copy_count == 0)
-		msg = "OK No messages copied.";
-	else {
-		i_assert(copy_count == uid2 - uid1 + 1);
+	else if (copy_count == 0) {
+		str_append(msg, "OK No messages copied.");
+		pool_unref(&changes.pool);
+	} else if (seq_range_count(&changes.saved_uids) == 0) {
+		/* not supported by backend (virtual) */
+		str_append(msg, "OK Copy completed.");
+		pool_unref(&changes.pool);
+	} else {
+		i_assert(copy_count == seq_range_count(&changes.saved_uids));
 
-		if (uid1 == 0)
-			msg = "OK Copy completed.";
-		else if (uid1 == uid2) {
-			msg = t_strdup_printf("OK [COPYUID %u %s %u] "
-					      "Copy completed.",
-					      uid_validity, src_uidset, uid1);
-		} else {
-			msg = t_strdup_printf("OK [COPYUID %u %s %u:%u] "
-					      "Copy completed.",
-					      uid_validity, src_uidset,
-					      uid1, uid2);
-		}
+		str_printfa(msg, "OK [COPYUID %u %s ", changes.uid_validity,
+			    src_uidset);
+		imap_write_seq_range(msg, &changes.saved_uids);
+		str_append(msg, "] Copy completed.");
+		pool_unref(&changes.pool);
 	}
 
+	dest_storage = mailbox_get_storage(destbox);
 	if (destbox != client->mailbox) {
 		sync_flags |= MAILBOX_SYNC_FLAG_FAST;
 		imap_flags |= IMAP_SYNC_FLAG_SAFE;
-		mailbox_close(&destbox);
+		mailbox_free(&destbox);
 	}
 
 	if (ret > 0)
-		return cmd_sync(cmd, sync_flags, imap_flags, msg);
+		return cmd_sync(cmd, sync_flags, imap_flags, str_c(msg));
 	else if (ret == 0) {
 		/* some messages were expunged, sync them */
 		return cmd_sync(cmd, 0, 0,
 			"NO ["IMAP_RESP_CODE_EXPUNGEISSUED"] "
 			"Some of the requested messages no longer exist.");
 	} else {
-		client_send_storage_error(cmd, storage);
+		client_send_storage_error(cmd, dest_storage);
 		return TRUE;
 	}
 }
