@@ -1,7 +1,7 @@
-/* Copyright (c) 2006-2010 Dovecot authors, see the included COPYING file */
+/* Copyright (c) 2006-2011 Dovecot authors, see the included COPYING file */
 
-#include "common.h"
-#include "commands.h"
+#include "imap-common.h"
+#include "imap-commands.h"
 
 bool cmd_x_cancel(struct client_command_context *cmd)
 {
