@@ -1,4 +1,4 @@
-/* Copyright (c) 2004-2010 Dovecot authors, see the included COPYING file */
+/* Copyright (c) 2004-2011 Dovecot authors, see the included COPYING file */
 
 #include "lib.h"
 #include "array.h"
@@ -76,16 +76,17 @@ static void keywords_ext_register(struct mail_index_sync_map_ctx *ctx,
 				  uint32_t ext_map_idx, uint32_t reset_id,
 				  uint32_t hdr_size, uint32_t keywords_count)
 {
-	buffer_t *ext_intro_buf;
+	buffer_t ext_intro_buf;
 	struct mail_transaction_ext_intro *u;
+	unsigned char ext_intro_data[sizeof(*u) +
+				     sizeof(MAIL_INDEX_EXT_KEYWORDS)-1];
 
 	i_assert(keywords_count > 0);
 
-	ext_intro_buf =
-		buffer_create_static_hard(pool_datastack_create(), sizeof(*u) +
-					  sizeof(MAIL_INDEX_EXT_KEYWORDS)-1);
+	buffer_create_data(&ext_intro_buf, ext_intro_data,
+			   sizeof(ext_intro_data));
 
-	u = buffer_append_space_unsafe(ext_intro_buf, sizeof(*u));
+	u = buffer_append_space_unsafe(&ext_intro_buf, sizeof(*u));
 	u->ext_id = ext_map_idx;
 	u->reset_id = reset_id;
 	u->hdr_size = hdr_size;
@@ -99,7 +100,7 @@ static void keywords_ext_register(struct mail_index_sync_map_ctx *ctx,
 
 	if (ext_map_idx == (uint32_t)-1) {
 		u->name_size = strlen(MAIL_INDEX_EXT_KEYWORDS);
-		buffer_append(ext_intro_buf, MAIL_INDEX_EXT_KEYWORDS,
+		buffer_append(&ext_intro_buf, MAIL_INDEX_EXT_KEYWORDS,
 			      u->name_size);
 	}
 
@@ -140,14 +141,16 @@ keywords_header_add(struct mail_index_sync_map_ctx *ctx,
 
 	if (buf == NULL) {
 		/* create new / replace broken header */
+		const unsigned int initial_keywords_count = 1;
+
 		buf = buffer_create_dynamic(pool_datastack_create(), 512);
 		kw_hdr = buffer_append_space_unsafe(buf, sizeof(*kw_hdr));
-		kw_hdr->keywords_count = 1;
+		kw_hdr->keywords_count = initial_keywords_count;
 
                 keywords_count = kw_hdr->keywords_count;
 		rec_offset = buf->used;
 		name_offset_root = rec_offset +
-			kw_hdr->keywords_count * sizeof(kw_rec);
+			initial_keywords_count * sizeof(kw_rec);
 		name_offset = 0;
 	}
 
