@@ -180,8 +180,7 @@ static int sdbox_mailbox_create_indexes(struct mailbox *box,
 	sdbox_update_header(mbox, trans, update);
 	if (new_trans != NULL) {
 		if (mail_index_transaction_commit(&new_trans) < 0) {
-			mail_storage_set_internal_error(box->storage);
-			mail_index_reset_error(box->index);
+			mail_storage_set_index_error(box);
 			return -1;
 		}
 	}
