@@ -31,9 +31,8 @@ static void gr_init(void)
 	long size;
 
 	if (grbuf == NULL) {
-		/* OpenBSD (up to 4.8 at least) reports too low value in
-		   sysconf() */
-#ifndef __OpenBSD__
+		/* OpenBSD up to 4.9 reports too low value in sysconf() */
+#if !defined(__OpenBSD__) || OpenBSD >= 201111
 		size = sysconf(_SC_GETGR_R_SIZE_MAX);
 		if (size < 0)
 #endif
