@@ -1,4 +1,4 @@
-/* Copyright (c) 2005-2010 Dovecot authors, see the included COPYING file */
+/* Copyright (c) 2005-2011 Dovecot authors, see the included COPYING file */
 
 #include "lib.h"
 #include "buffer.h"
@@ -105,7 +105,8 @@ static int copy_to_temp_file(struct seekable_istream *sstream)
 
 		if (i_stream_read(sstream->fd_input) <= 0) {
 			i_error("istream-seekable: Couldn't read back "
-				"in-memory input");
+				"in-memory input %s",
+				i_stream_get_name(&stream->istream));
 			i_stream_destroy(&sstream->fd_input);
 			return -1;
 		}
@@ -131,7 +132,7 @@ static ssize_t read_more(struct seekable_istream *sstream)
 			/* full / error */
 			sstream->istream.istream.stream_errno =
 				sstream->cur_input->stream_errno;
-			return ret;
+			return -1;
 		}
 
 		/* go to next stream */
@@ -156,8 +157,9 @@ static bool read_from_buffer(struct seekable_istream *sstream, ssize_t *ret_r)
 	const unsigned char *data;
 	size_t size, pos, offset;
 
-	if (stream->istream.v_offset +
-	    (stream->pos - stream->skip) >= sstream->buffer->used) {
+	i_assert(stream->skip == 0);
+
+	if (stream->istream.v_offset + stream->pos >= sstream->buffer->used) {
 		/* need to read more */
 		if (sstream->buffer->used >= stream->max_buffer_size)
 			return FALSE;
@@ -175,6 +177,7 @@ static bool read_from_buffer(struct seekable_istream *sstream, ssize_t *ret_r)
 
 		/* we should have more now. */
 		data = i_stream_get_data(sstream->cur_input, &size);
+		i_assert(size > 0);
 		buffer_append(sstream->buffer, data, size);
 		i_stream_skip(sstream->cur_input, size);
 	}
@@ -360,6 +363,7 @@ i_stream_create_seekable(struct istream *input[],
 	sstream->context = context;
 	sstream->buffer = buffer_create_dynamic(default_pool, BUF_INITIAL_SIZE);
         sstream->istream.max_buffer_size = max_buffer_size;
+	sstream->fd = -1;
 
 	sstream->input = i_new(struct istream *, count + 1);
 	memcpy(sstream->input, input, sizeof(*input) * count);
